/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.
   
   GFX.C - Graphics routines.
*/

/* the includes */
#include <math.h>
#include "deu.h"

#include "zethdat.h"
extern DATAFILE *zethdat;

BITMAP *buffer;
BITMAP *scr_backup;
BITMAP *icon_bmp;

BYTE DPal[768];
int Palette[PALETTE_SIZE];

FONT *DethFont;
FONT *SmallFont;


/* the global variables */
double OrigX;			/* the X origin */
double OrigY;			/* the Y origin */
double Scale;			/* the scale value */
int PointerX;			/* X position of pointer */
int PointerY;			/* Y position of pointer */
int ScrMaxX;			/* maximum X screen coord */
int ScrMaxY;			/* maximum Y screen coord */
int ScrCenterX;			/* X coord of screen center */
int ScrCenterY;			/* Y coord of screen center */
Bool CustomLine;
int text_center = FALSE;
int text_color = 0;
int bg_color = -1;
int line_color = 0;
int line_width = 1;
int line_dotted = FALSE;
int Fullscreen;

static int res[7][2] = {{0,0},{0,0},{640,480},{800,600},
                        {1024,768},{1280,960},{1600,1200}};
static int kb_inited = FALSE;

/* initialise the graphics display */
void InitGfx(void)
{
	static int depths[] = { -1, 32, 24, 16, 15, 8, 0 };
	int buf_hack = FALSE;
	int card;
	int i;

	if(depths[0] < 0)
	{
		depths[0] = desktop_color_depth();
		if(depths[0] == 0)
		{
			for(i = 0;depths[i+1];++i)
				depths[i] = depths[i+1];
		}
	}

	if (VideoMode < 2 || VideoMode > 6)
		VideoMode = 2;

	// [CR] If windowed, use the desktop color depth(it's faster)
	if(!Fullscreen)
		card = GFX_AUTODETECT_WINDOWED;
	else
		card = GFX_AUTODETECT_FULLSCREEN;

	if(gfx_driver)
	{
		if(ScrMaxX == res[VideoMode][0] && ScrMaxY == res[VideoMode][1])
		{
			if((!gfx_driver->windowed && Fullscreen) ||
			   (gfx_driver->windowed && !Fullscreen))
				return;
		}
	}

	if(buffer && buffer == screen)
		buf_hack = TRUE;

	for(i = 0;depths[i];++i)
	{
		set_color_depth(depths[i]);
		if(set_gfx_mode(card, res[VideoMode][0], res[VideoMode][1], 0, 0) == 0)
			break;
	}

	if(depths[i] == 0)
	{
		Fullscreen = !Fullscreen;
		if(!Fullscreen)
			card = GFX_AUTODETECT_WINDOWED;
		else
			card = GFX_AUTODETECT_FULLSCREEN;

		for(i = 0;depths[i];++i)
		{
			set_color_depth(depths[i]);
			if(set_gfx_mode(card, res[VideoMode][0], res[VideoMode][1], 0, 0) == 0)
				break;
		}

		if(depths[i] == 0)
			ProgError("Graphics error: %s\n", allegro_error);/*(in case you are running this from a command "
					  "prompt you might have to switch to fullscreen)", allegro_error);*/
	}

	if(!kb_inited)
	{
		if(install_keyboard() != 0)
		{
			TermGfx();
			allegro_message("Error installing keyboard handler!\n");
			exit(1);
		}
		kb_inited = TRUE;
	}

	if(gfx_driver->windowed)
		set_display_switch_mode(SWITCH_BACKGROUND);
	else
		set_display_switch_mode(SWITCH_BACKAMNESIA);

	atexit(TermGfx);

	SetDoomPalette(0);

	clear_bitmap(screen);

	if(buffer)
		destroy_bitmap(buffer);
	buffer = create_bitmap(SCREEN_W, SCREEN_H);

	if(icon_bmp)
		destroy_bitmap(icon_bmp);
	icon_bmp = create_bitmap(32, 32);

	scr_backup = screen;

	if(buf_hack)
		screen = buffer;

    ScrMaxX = SCREEN_W;
    ScrMaxY = SCREEN_H;
    ScrCenterX = ScrMaxX / 2;
    ScrCenterY = ScrMaxY / 2;

	DethFont = (FONT*)zethdat[DETHFONT].dat;
	SmallFont = (FONT*)zethdat[SMALLFNT].dat;

	CheckMouseDriver();
	gui_fg_color = TranslateToGameColor(BLACK);
	gui_mg_color = TranslateToGameColor(DARKGRAY);
	gui_bg_color = TranslateToGameColor(WHITE);
}

/* terminate the graphics display */
void TermGfx(void)
{
	// [CR] We don't want to quit from a graphics mode if there's no console
	//      to fall to
#ifdef ALLEGRO_CONSOLE_OK
	if(gfx_driver && !gfx_driver->windowed)
	{
		if(kb_inited)
		{
			remove_keyboard();
			kb_inited = FALSE;
		}

		if(screen == buffer)
			screen = scr_backup;

		set_gfx_mode(GFX_TEXT, 0, 0, 0, 0);
	}
#endif
}


/* clear the map screen */
/*AJB*/
void ClearMapScreen(int top, int maxy)
{
	HideMousePointer();
	rectfill(screen, 0, top, ScrMaxX-1, ScrMaxY - maxy, TranslateToGameColor(BLACK));
	ShowMousePointer();
}

/* set the current drawing color */
void SetColor(int color)
{
	text_color = TranslateToGameColor(color);
	line_color = text_color;
}

/* Set the text background color (-1 is none) */
void SetTextBg(int color)
{
	if (color < 0)
		bg_color = -1;
	else
		bg_color = TranslateToGameColor(color);
}

/*
	Enable or disable horizontal text centering
 */
void SetTextCentered(Bool center)
{
	text_center = center;
}

/*
	set the write mode
 */
void SetWriteMode(int mode)
{
	switch (mode)
	{
		case COPY_PUT:
			mode = DRAW_MODE_SOLID;
			break;
		case XOR_PUT:
			mode = DRAW_MODE_XOR;
			break;
		default:
			return;
	}

	drawing_mode(mode, screen, 0, 0);
}

void SetLineDotted(Bool dotted)
{
	line_dotted = dotted;
}

void SetLineThickness(int width)
{
	line_width = width;
}

/* draw a line on the screen from map coords */
void DrawMapLine(int mapXstart, int mapYstart, int mapXend, int mapYend)
{
/*	int Xstart = SCREENX(mapXstart);
	int Ystart = SCREENY(mapYstart);
	int Xend = SCREENX(mapXend);
	int Yend = SCREENY(mapYend);

	DrawScreenLine(Xstart, Ystart, Xend, Yend);*/
	DrawScreenLine(SCREENX(mapXstart), SCREENY(mapYstart), SCREENX(mapXend), SCREENY(mapYend));
}



/* draw a circle on the screen from map coords */
void DrawMapCircle(int mapXcenter, int mapYcenter, int mapRadius)
{
	int x = SCREENX(mapXcenter);
	int y = SCREENY(mapYcenter);
	int r = (int)(mapRadius * Scale);

	if(x-r >= ScrMaxX || x+r < 0 || y-r >= ScrMaxY || y+r < 0)
		return;

	HideMousePointer();
	circle(screen, x, y, r, text_color);
	ShowMousePointer();
}



/* draw an arrow on the screen from map coords */
void DrawMapVector(int mapXstart, int mapYstart, int mapXend, int mapYend)
{
	int  scrXstart = SCREENX(mapXstart);
	int  scrYstart = SCREENY(mapYstart);
	int  scrXend   = SCREENX(mapXend);
	int  scrYend   = SCREENY(mapYend);
	double r       = hypot((double)(scrXstart - scrXend), (double)(scrYstart - scrYend));
	int  scrXoff   = (r >= 1.0) ? (int)((scrXstart - scrXend) * 4.0 / r * Scale) : 0;
	int  scrYoff   = (r >= 1.0) ? (int)((scrYstart - scrYend) * 4.0 / r * Scale) : 0;

	HideMousePointer();
	DrawScreenLine(scrXstart, scrYstart, scrXend, scrYend);

	scrXstart = scrXend + 2 * scrXoff;
	scrYstart = scrYend + 2 * scrYoff;
	DrawScreenLine(scrXstart - scrYoff, scrYstart + scrXoff, scrXend, scrYend);
	DrawScreenLine(scrXstart + scrYoff, scrYstart - scrXoff, scrXend, scrYend);
	ShowMousePointer();
}



/* draw an arrow on the screen from map coords and angle (0 - 65535) */
void DrawMapArrow(int mapXstart, int mapYstart, UBCINT angle)
{
	int  mapXend   = mapXstart + (int)(50 * cos(angle / 10430.37835));
	int  mapYend   = mapYstart + (int)(50 * sin(angle / 10430.37835));
	int  scrXstart = SCREENX(mapXstart);
	int  scrYstart = SCREENY(mapYstart);
	int  scrXend   = SCREENX(mapXend);
	int  scrYend   = SCREENY(mapYend);
	double r       = hypot(scrXstart - scrXend, scrYstart - scrYend);
	int  scrXoff   = (r >= 1.0) ? (int)((scrXstart - scrXend) * 4.0 / r * Scale) : 0;
	int  scrYoff   = (r >= 1.0) ? (int)((scrYstart - scrYend) * 4.0 / r * Scale) : 0;

//	if(scrXstart >= ScrMaxX || scrXend < 0 || scrYstart >= ScrMaxY || scrYend < 0)
//		return;

	HideMousePointer();
	DrawScreenLine(scrXstart, scrYstart, scrXend, scrYend);
	scrXstart = scrXend + 2 * scrXoff;
	scrYstart = scrYend + 2 * scrYoff;
	DrawScreenLine(scrXstart - scrYoff, scrYstart + scrXoff, scrXend, scrYend);
	DrawScreenLine(scrXstart + scrYoff, scrYstart - scrXoff, scrXend, scrYend);
	ShowMousePointer();
}

/* draw a filled box on the map */
void DrawMapBox(int x1, int y1, int x2, int y2)
{
	x1 = SCREENX(x1);
	x2 = SCREENX(x2);
	y1 = SCREENY(y1);
	y2 = SCREENY(y2);

	if(x1 >= ScrMaxX || x2 < 0 || y1 >= ScrMaxY || y2 < 0)
		return;

	HideMousePointer();
	rectfill(screen, x1, y1, x2, y2, line_color);
	ShowMousePointer();
}

/* draw a filled box on the map with a black frame. */
void DrawMapTBox(int x1, int y1, int x2, int y2)
{
	x1 = SCREENX(x1);
	x2 = SCREENX(x2);
	y1 = SCREENY(y1);
	y2 = SCREENY(y2);

	if(x1 >= ScrMaxX || x2 < 0 || y1 >= ScrMaxY || y2 < 0)
		return;

	HideMousePointer();
	rectfill(screen, x1, y1, x2, y2, line_color);
	rect(screen, x1, y1, x2, y2, TranslateToGameColor(BLACK));
	ShowMousePointer();
}

/* draw a hollow box on the map */
void DrawMapHollowBox(int x1, int y1, int x2, int y2)
{
	int w;

	x1 = SCREENX (x1);
	x2 = SCREENX (x2);
	y1 = SCREENY (y1);
	y2 = SCREENY (y2);

	if(x1 >= ScrMaxX || x2 < 0 || y1 >= ScrMaxY || y2 < 0)
		return;

	HideMousePointer();
	w = line_width;
	do {
		rect(screen, x1, y1, x2, y2, line_color);

		if(--w == 0)
			break;

		++x1; ++y1;
		--x2; --y2;
	} while(1);
	ShowMousePointer();
}

/* draw a line on the screen from screen coords */
void dotted_line(BITMAP *bmp, int x, int y, int d)
{
	static int t = 0;

	if(!((t++)&3))
		putpixel(bmp, x, y, d);
}

int xoff, yoff;
void wide_line(BITMAP *bmp, int x, int y, int d)
{
	line(bmp, x-xoff, y-yoff, x+xoff, y+yoff, d);
}

void FatLine(BITMAP *bmp, int x1, int y1, int x2, int y2, int fatness, int colour)
{
	int dx = x2-x1, dy = y2-y1;
	double px, py, length;

	// Special cases
	if(!dx)
	{
		int fat = fatness / 2;
		rectfill(bmp, x1-fat, y1, x2+fat, y2, colour);
		return;
	}
	if(!dy)
	{
		int fat = fatness / 2;
		rectfill(bmp, x1, y1-fat, x2, y2+fat, colour);
		return;
	}

	// Sus out length
	length = hypot(dy, dx);//sqrt(dx*dx + dy*dy);

	px = (dy / length) * fatness;
	py = (dx / length) * fatness;

	// [CR] This seems to work better
	xoff = -px;
	yoff =  py;
	do_line(bmp, x1, y1, x2, y2, colour, wide_line);
}

void DrawScreenLine(int Xstart, int Ystart, int Xend, int Yend)
{
	if(Xstart <= Xend)
	{
		if(Xstart >= ScrMaxX || Xend < 0)
			return;
	}
	else
	{
		if(Xstart < 0 || Xend >= ScrMaxX)
			return;
	}

	if(Ystart <= Yend)
	{
		if(Ystart >= ScrMaxY || Yend < 0)
			return;
	}
	else
	{
		if(Ystart < 0 || Yend >= ScrMaxY)
			return;
	}

	HideMousePointer();
	if(line_width > 1)
		FatLine(screen, Xstart, Ystart, Xend, Yend, line_width-1, line_color);
	else if(line_dotted)
		do_line(screen, Xstart, Ystart, Xend, Yend, line_color, dotted_line);
	else
		line(screen, Xstart, Ystart, Xend, Yend, line_color);
	ShowMousePointer();
}

/* draw a filled in box on the screen from screen coords */
void DrawScreenBox(int Xstart, int Ystart, int Xend, int Yend)
{
	HideMousePointer();
	rectfill(screen, Xstart, Ystart, Xend, Yend, text_color);
	ShowMousePointer();
}

void DrawScreenFrame(int Xstart, int Ystart, int Xend, int Yend)
{
	HideMousePointer();
	rect(screen, Xstart, Ystart, Xend, Yend, text_color);
	ShowMousePointer();
}


/* draw a filled-in 3D-box on the screen from screen coords */
void DrawScreenBox3D(int Xstart, int Ystart, int Xend, int Yend)
{
	HideMousePointer();
	rectfill(screen, Xstart + 1, Ystart + 1, Xend - 1, Yend - 1,
	         TranslateToGameColor(DARKGRAY));
    SetColor(DARKERGRAY);
    line(screen, Xstart, Yend, Xend, Yend, line_color);
    line(screen, Xend, Ystart, Xend, Yend, line_color);
    SetColor(LIGHTGRAY);
    line(screen, Xstart, Ystart, Xend, Ystart, line_color);
    line(screen, Xstart, Ystart, Xstart, Yend, line_color);
    SetColor(BLACK);
	ShowMousePointer();
}



/* draw a hollow 3D-box on the screen from screen coords */
void DrawScreenBoxHollow(int Xstart, int Ystart, int Xend, int Yend)
{
	HideMousePointer();
	rectfill(screen, Xstart + 1, Ystart + 1, Xend - 1, Yend - 1,
	         TranslateToGameColor( BLACK));
    SetColor(LIGHTGRAY);
    line(screen, Xstart, Yend, Xend, Yend, line_color);
    line(screen, Xend, Ystart, Xend, Yend, line_color);
    SetColor(DARKERGRAY);
    line(screen, Xstart, Ystart, Xend, Ystart, line_color);
    line(screen, Xstart, Ystart, Xstart, Yend, line_color);
    SetColor(LIGHTGRAY);
	ShowMousePointer();
}



void DrawScreenBar(int Xstart, int Ystart, int Width, Bool Reversed)
{
	HideMousePointer();
	SetColor(Reversed ? DARKERGRAY : LIGHTGRAY);
    line(screen, Xstart, Ystart, Xstart + Width, Ystart, line_color);
	SetColor(Reversed ? LIGHTGRAY : DARKERGRAY);
    line(screen, Xstart, Ystart + 1, Xstart + Width, Ystart + 1, line_color);
	ShowMousePointer();
}

// [RH] Why did the buttons have an illegible bright shadow?
void DrawScreenButton( int Xstart, int Ystart, int Width, int Height, char *Text, Bool Filled)
{
	HideMousePointer();
	if (Filled == TRUE) {
		rectfill(screen, Xstart + 1, Ystart + 1, Xstart + Width - 1,
		         Ystart + Height - 1, TranslateToGameColor(DARKGRAY));
   	}
    SetColor(DARKERGRAY);
    line(screen, Xstart, Ystart + Height, Xstart + Width, Ystart + Height, line_color);
    line(screen, Xstart + Width, Ystart, Xstart + Width, Ystart + Height, line_color);
	SetColor(LIGHTGRAY);
	line(screen, Xstart, Ystart, Xstart + Width, Ystart, line_color);
	line(screen, Xstart, Ystart, Xstart, Ystart + Height, line_color); 
	ScreenBoxArea(Xstart - 1, Ystart - 1, Width + 2, Height + 2, BLACK);
	if(Text != NULL)
	{
		SetColor(BLACK);
		textout_centre_ex(screen, SmallFont, Text, Xstart + (Width >> 1),
		                  Ystart + (Height >> 1) - 6 + 1, text_color, bg_color);
	}
	ShowMousePointer();
}

void DrawScreenButtonGray(int Xstart, int Ystart, int Width, int Height, char *Text, Bool Filled)
{
	HideMousePointer();
	if (Filled == TRUE)
	{
		rectfill(screen, Xstart + 1, Ystart + 1, Xstart + Width - 1,
		         Ystart + Height - 1, TranslateToGameColor(DARKGRAY));
   	}
    SetColor(DARKERGRAY);
    line(screen, Xstart, Ystart + Height, Xstart + Width, Ystart + Height, line_color);
    line(screen, Xstart + Width, Ystart, Xstart + Width, Ystart + Height, line_color); 
	SetColor(LIGHTGRAY);
	line(screen, Xstart, Ystart, Xstart + Width, Ystart, line_color);
	line(screen, Xstart, Ystart, Xstart, Ystart + Height, line_color);
	ScreenBoxArea(Xstart - 1, Ystart - 1, Width + 2, Height + 2, BLACK);
	if (Text != NULL)
	{
		SetColor (DARKERGRAY);
		textout_centre_ex(screen, SmallFont, Text, Xstart + (Width >> 1),
		                  Ystart + (Height >> 1) - 6 + 1, text_color, bg_color);
	}
	ShowMousePointer();
}

void DrawScreenButtonIn( int Xstart, int Ystart, int Width, int Height, char *Text, Bool Filled)
{
	HideMousePointer();
	if (Filled)
	{
		rectfill(screen, Xstart + 1, Ystart + 1, Xstart + Width - 1,
		         Ystart + Height - 1, TranslateToGameColor(DARKGRAY));
   	}
	SetColor(LIGHTGRAY);
	line(screen, Xstart, Ystart + Height, Xstart + Width, Ystart + Height, line_color);
	line(screen, Xstart + Width, Ystart, Xstart + Width, Ystart + Height, line_color); 
	SetColor(DARKERGRAY);
	line(screen, Xstart, Ystart, Xstart + Width, Ystart, line_color);
	line(screen, Xstart, Ystart, Xstart, Ystart + Height, line_color); 
	if (Text != NULL)
	{
		SetColor (BLACK);
		textout_centre_ex(screen, SmallFont, Text, Xstart + (Width >> 1),
		                  Ystart + (Height >> 1) - 6 + 1, text_color, bg_color);
	}
	ShowMousePointer();
}

void DrawScreenButtonOut(int Xstart, int Ystart, int Width, int Height, Bool Filled)
{
	HideMousePointer();
	if (Filled)
	{
		rectfill(screen, Xstart + 1, Ystart + 1, Xstart + Width - 1,
		         Ystart + Height - 1, TranslateToGameColor(BLUE));
   	}
	SetColor(DARKERGRAY);
	line(screen, Xstart, Ystart + Height, Xstart + Width, Ystart + Height, line_color);
	line(screen, Xstart + Width, Ystart, Xstart + Width, Ystart + Height, line_color); 
	SetColor(LIGHTGRAY);
	line(screen, Xstart, Ystart, Xstart + Width, Ystart, line_color);
	line(screen, Xstart, Ystart, Xstart, Ystart + Height, line_color); 
	ShowMousePointer();
}


void DrawScreenWindow(int x, int y, int width, int height, char *title, int color)
{
	int dx = x + width;
	int dy = y + height;

	HideMousePointer();
	SetColor(LIGHTGRAY);
	line(screen, x, y, dx - 1, y, line_color);
	line(screen, x, y, x, dy - 1, line_color);
	SetColor(DARKERGRAY);
	line(screen, x, dy, dx , dy, line_color);
	line(screen, dx, y, dx, dy, line_color);
	rectfill(screen, x + 1, y + 1, dx - 1, dy - 1, TranslateToGameColor(DARKGRAY));
	rectfill(screen, x + 3, y + 3, dx - 3, y + 17, TranslateToGameColor(color));
	SetColor(BLACK);
	DrawScreenTextSmall(x + 6, y + 4, "%s", title);
	ShowMousePointer();
}

void ScreenHighlightArea(int x, int y, int width, int height, char *text, int color, int textcolor)
{
	int dx = x + width;
	int dy = y + height;

	HideMousePointer();
	rectfill(screen, x + 1, y + 1, dx - 1, dy - 1, TranslateToGameColor(color));
	SetColor(textcolor);
	DrawScreenText(x + 3, y + 2, "%s", text);
	ShowMousePointer();
}

void ScreenBoxArea(int x, int y, int width, int height, int color)
{
	int dx = x + width;
	int dy = y + height;

	SetColor(color);

	HideMousePointer();
    line(screen, x, y, dx, y, line_color);
    line(screen, x, y, x, dy, line_color);
    line(screen, dx, y, dx, dy, line_color);
    line(screen, x, dy, dx, dy, line_color);
	ShowMousePointer();
}

/* draw a meter bar on the screen from screen coords (in a hollow box); max. value = 1.0 */
void DrawScreenMeter( int Xstart, int Ystart, int Xend, int Yend, float value)
{
    if (value < 0.0)
		value = 0.0;
    if (value > 1.0)
		value = 1.0;

	HideMousePointer();
	rectfill(screen, Xstart + 1 + (int) ((Xend - Xstart - 2) * value), Ystart + 1,
	         Xend - 1, Yend - 1, TranslateToGameColor(BLACK));
	rectfill(screen, Xstart + 1, Ystart + 1, Xstart + 1 + (int) ((Xend - Xstart - 2) * value), Yend - 1, TranslateToGameColor( LIGHTGREEN));
	ShowMousePointer();
}


/*
   write text to the screen
   */

void DrawScreenText( int Xstart, int Ystart, char *msg, ...)
{
    static int lastX;
    static int lastY;
    char temp[256] = "";
	int len;
    va_list args;

    va_start( args, msg);
    len = vsprintf( temp, msg, args);
    va_end( args);

    if (Xstart == -1)
		Xstart = lastX;
    if (Ystart == -1)
		Ystart = lastY + 10;

	HideMousePointer();

	if(!text_center)
		textout_ex(screen, DethFont, temp, Xstart, Ystart+1, text_color, bg_color);
	else
		textout_centre_ex(screen, DethFont, temp, Xstart, Ystart+1, text_color, bg_color);

	ShowMousePointer();

	lastX = Xstart;
	lastY = Ystart;
}

void DrawScreenTextSmall(int Xstart, int Ystart, char *msg, ...)
{
    static int lastX;
    static int lastY;
    char temp[ 120];
	int len;

    va_list args;
    va_start(args, msg);
    len = vsprintf(temp, msg, args);
    va_end( args);

    if (Xstart <= -1)
		Xstart = lastX;
    if (Ystart <= -1)
		Ystart = lastY + 10;

	HideMousePointer();

	if(!text_center)
		textout_ex(screen, SmallFont, temp, Xstart, Ystart, text_color, bg_color);
	else
		textout_centre_ex(screen, SmallFont, temp, Xstart, Ystart, text_color, bg_color);

	ShowMousePointer();

	lastX = Xstart;
	lastY = Ystart;
}

void DrawScreen3DText( int Xstart, int Ystart, int col, int scol, char *msg, ...)
{
    static int lastX;
    static int lastY; 
    char temp[ 120];
	int len;

    va_list args;
    va_start( args, msg);
    len = vsprintf( temp, msg, args);
    va_end( args);

    if (Xstart <= -1)
		Xstart = lastX;
    if (Ystart <= -1)
		Ystart = lastY + 10;

	SetColor(scol);
	HideMousePointer();
	if(!text_center)
	{
		textout_ex(screen, DethFont, temp, Xstart, Ystart, text_color, bg_color);
		SetColor(col);
		textout_ex(screen, DethFont, temp, Xstart-1, Ystart-1, text_color, bg_color);
	}
	else
	{
		textout_centre_ex(screen, DethFont, temp, Xstart, Ystart, text_color, bg_color);
		SetColor(col);
		textout_centre_ex(screen, DethFont, temp, Xstart-1, Ystart-1, text_color, bg_color);
	}
	ShowMousePointer();

	lastX = Xstart;
	lastY = Ystart;
}


/*
   draw (or erase) the pointer if we aren't using the mouse
   */

void DrawPointer()
{
	/* use XOR mode : drawing the pointer twice erases it */
}



/*
   load one "playpal" palette and change all palette colours
   */

void SetDoomPalette( int playpalnum)
{
	PALETTE pal;
	BYTE	gamma[256];
	MDirPtr	dir;
	int	n;

	if (playpalnum < 0 || playpalnum > 13)
		return;
	dir = FindMasterDir( MasterDir, "PLAYPAL");

	if (dir)
	{
		BasicWadSeek( dir->wadfile, dir->dir.start);
		for (n = 0; n <= playpalnum; n++)
			BasicWadRead( dir->wadfile, DPal, 768L);

		if (Gamma > 0 && Gamma <= 4) 
		{
			static const double gmult[5] = {1.0, 0.75, 0.55, 0.4, 0.3};
			double mult = gmult[Gamma];

			for (n = 0; n < 256; ++n)
			{
				gamma[n] = pow (((double)n / 255.0), mult) * 255.0;
			}
		}
		else
		{
			for (n = 0; n < 256; ++n)
			{
				gamma[n] = n;
			}
		}

		for (n = 0; n < 256; n++)
		{
			pal[n].r = gamma[DPal[3*n]] >> 2;
			pal[n].g = gamma[DPal[3*n+1]] >> 2;
			pal[n].b = gamma[DPal[3*n+2]] >> 2;
		}
		set_palette(pal);
	}
}

/*
   translate a standard color to Doom palette 0 (approx.)
   */

int TranslateToGameColor(int color)
{
	int c = palette_color[Palette[color]];

	if(c == bitmap_mask_color(screen))
	{
		if(c == 0)
			return makecol(0, 0, 0);
		else
			--c;
	}

	return c;
}


/*
   translate (dx, dy) into an integer angle value (0-65535)
   */

UBCINT ComputeAngle( int dx, int dy)
{
    return (UBCINT) (atan2( (double) dy, (double) dx) * 10430.37835 + 0.5);
    /* Yes, I know this function could be in another file, but */
    /* this is the only source file that includes <math.h>...  */
}



/*
   compute the distance from (0, 0) to (dx, dy)
   */

UBCINT ComputeDist( int dx, int dy)
{
    return (UBCINT) (hypot( (double) dx, (double) dy) + 0.5);
    /* Yes, I know this function could be in another file, but */
    /* this is the only source file that includes <math.h>...  */
}



/*
   insert the vertices of a new polygon
   */

void InsertPolygonVertices( int centerx, int centery, int sides, int radius)
{
    int n;
    
    for (n = 0; n < sides; n++)
	{
		InsertObject( OBJ_VERTEXES, -1,
			centerx + (int) ((double) radius * cos( 2*M_PI * (double) n / (double) sides)),
			centery + (int) ((double) radius * sin( 2*M_PI * (double) n / (double) sides)));
	}
    /* Yes, I know... etc. */
}



/*
   move (x, y) to a new position: rotate and scale around (0, 0)
   */

//void RotateAndScaleCoords(int *x, int *y, double angle, double scale)
//{
//    double r, theta;
//    
//    r = hypot((double) *x, (double) *y);
//    theta = atan2((double) *y, (double) *x);
//    *x = (int) (r * scale * cos(theta + angle) + 0.5);
//    *y = (int) (r * scale * sin(theta + angle) + 0.5);
//}
// jff
// use double throughout, round final answer for increased accuracy
//
void RotateAndScaleCoords(double *x, double *y, double angle, double scale)
{
    double r, theta;
    
    r = hypot(*x,*y);
    theta = atan2(*y,*x);
    *x = r * scale * cos(theta + angle);
    *y = r * scale * sin(theta + angle);
}

/* end of file */
