/*
   Mouse interface by Raphal Quinet <quinet@montefiore.ulg.ac.be>
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.
   
   MOUSE.C - Mouse driver routines.
   */

/* the includes */
#include "deu.h"

#include "zethdat.h"
extern DATAFILE *zethdat;

// is there a mouse driver?
static Bool UseMouse;

// so we don't show the mouse durring drawing operations prematurely
int mouse_hide_level = 0;

// Mouse graphic data. 0 = transparent, 1 = black, 2 = white
static const unsigned char m_data[16][12] = {
	{ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 1, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 1, 2, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 1, 2, 2, 2, 1, 0, 0, 0, 0, 0, 0, 0 },
	{ 1, 2, 2, 2, 2, 1, 0, 0, 0, 0, 0, 0 },
	{ 1, 2, 2, 2, 2, 2, 1, 0, 0, 0, 0, 0 },
	{ 1, 2, 2, 2, 2, 2, 2, 1, 0, 0, 0, 0 },
	{ 1, 2, 2, 2, 2, 2, 2, 2, 1, 0, 0, 0 },
	{ 1, 2, 2, 2, 2, 2, 2, 2, 2, 1, 0, 0 },
	{ 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1, 0 },
	{ 1, 2, 2, 2, 2, 2, 2, 2, 2, 2, 2, 1 },
	{ 1, 2, 2, 2, 2, 1, 1, 1, 1, 1, 1, 0 },
	{ 1, 2, 2, 2, 1, 0, 0, 0, 0, 0, 0, 0 },
	{ 1, 2, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 1, 2, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0 },
	{ 0, 1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0 }
};

/* The mouse bitmap */
BITMAP *mouse_bmp = NULL;

extern BITMAP *scr_backup;

/* initialize the mouse driver */
void CheckMouseDriver()
{
	if(install_mouse() != -1)
	{
		int x, y, w, b, m;

		// [CR] Clear the current mouse sprite, if any
		set_mouse_sprite(NULL);
		if(mouse_bmp)
		{
			destroy_bitmap(mouse_bmp);
			mouse_bmp = NULL;
		}
		// [CR] Get the white, black, and transparent color for the
		//      current display depth
		w = makecol(255, 255, 255);
		b = makecol(0, 0, 0);
		m = bitmap_mask_color(screen);

		// [CR] Create the mouse bitmap, using the data provided above to
		//      make it
		mouse_bmp = create_bitmap(12, 16);
		if(mouse_bmp)
		{
			clear_to_color(mouse_bmp, m);

			for(y = 0;y < mouse_bmp->h;++y)
			{
				for(x = 0;x < mouse_bmp->w;++x)
				{
					switch(m_data[y][x])
					{
					case 1:
						putpixel(mouse_bmp, x, y, b);
						break;
					case 2:
						putpixel(mouse_bmp, x, y, w);
						break;
					}
				}
			}
		}
		set_mouse_sprite(mouse_bmp);

		mouse_hide_level = 0;
		show_mouse(scr_backup);

		UseMouse = TRUE;
	}
	else
	{
		UseMouse = FALSE;
		alert("Error!", "Mouse init failed!", "", "Ok", NULL, 0, 0);
	}
}


void ShowMousePointer()
{
	if (!UseMouse)
		return;

	if(mouse_hide_level)
	{
		if(--mouse_hide_level == 0)
			show_mouse(screen);
	}
}



void HideMousePointer()
{
	if (!UseMouse)
		return;

	if(mouse_hide_level == 0)
		show_mouse(NULL);

	++mouse_hide_level;
}


void GetMouseCoords(int *x, int *y, int *buttons)
{
	if (!UseMouse)
	{
		*x = ScrCenterX;
		*y = ScrCenterY;
		*buttons = 0;
		return;
	}

	if(mouse_needs_poll())
		poll_mouse();

	*x = mouse_x;
	*y = mouse_y;
	*buttons = mouse_b;
}



void SetMouseCoords( BCINT x, BCINT y)
{
	if (!UseMouse)
		return;

	if(gfx_driver && !gfx_driver->windowed)
		position_mouse(x, y);
}



void SetMouseLimits( BCINT x0, BCINT y0, BCINT x1, BCINT y1)
{
	if(!UseMouse)
		return;

	set_mouse_range(x0,y0,x1,y1);
}


Bool MouseInArea(BCINT x, BCINT y, BCINT length, BCINT height)
{
	int PointerX, PointerY, buttons;

	GetMouseCoords(&PointerX, &PointerY, &buttons);
	return (PointerX >= x && PointerX < x + length && PointerY >= y && PointerY < y + height);
}

Bool MouseClickedArea(BCINT x, BCINT y, BCINT length, BCINT height)
{
	int PointerX, PointerY, buttons;

	if (!UseMouse)
		return FALSE;

	GetMouseCoords(&PointerX, &PointerY, &buttons);
	if (PointerX >= x && PointerX < x + length && PointerY >= y && PointerY < y + height)
	{
		while (buttons & 1)
		{
			yield();
			GetMouseCoords( &PointerX, &PointerY, &buttons);
		}
	}

	if (!(buttons & 1) && PointerX >= x && PointerX < x + length && PointerY >= y && PointerY < y + height)
		return TRUE;

	return FALSE;
}

void ResetMouseLimits()
{
	if (!UseMouse)
		return;

    set_mouse_range(0,0,ScrMaxX,ScrMaxY);
}

void MouseSetColors(BCINT color)
{
	int x, y, c;

	if(!UseMouse)
		return;

	c = TranslateToGameColor(color);

	for(y = 0;y < mouse_bmp->h;++y)
	{
		for(x = 0;x < mouse_bmp->w;++x)
		{
			if(m_data[y][x] == 2)
				putpixel(mouse_bmp, x, y, c);
		}
	}
}
/* end of file */
