/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.

   STRUCT_H.H - HEXEN WAD files data structures.
   */


/*
   this data structure contains the information about the THINGS
   */

/* protection against multiple includes */
#ifndef _WSTRUCTS_H
#define _WSTRUCTS_H

struct Thing
{
    short tid;
    short xpos;     /* x position */
    short ypos;     /* y position */
    short zpos;	    /* z position (height) */
    short angle;    /* facing angle */
    short type;     /* thing type */
    short flags;    /* appears when? */
    BYTE special; 	/* special command to perform */
    BYTE arg[5];	/* special arguments */

	short nextByID;
	short prevByID;
};
typedef struct Thing *TPtr;



/*
   this data structure contains the information about the LINEDEFS
   */
struct DoomLineDef
{
    short start;     /* from this vertex ... */
    short end;       /* ... to this vertex */
    short flags;     /* see NAMES.C for more info */
    short tag;
    short special;      /* see NAMES.C for more info */
    short sidedef1;  /* sidedef */
    short sidedef2;  /* only if this line adjoins 2 sectors */
};
typedef struct DoomLineDef *DLDPtr;

struct LineDef
{
    short start;     /* from this vertex ... */
    short end;       /* ... to this vertex */
    short flags;     /* see NAMES.C for more info */
    BYTE special;      /* see NAMES.C for more info */
	BYTE arg[5];
    short sidedef1;  /* sidedef */
    short sidedef2;  /* only if this line adjoins 2 sectors */
};
typedef struct LineDef *LDPtr;

#define ML_BLOCKING			0x0001	// solid, is an obstacle
#define ML_BLOCKMONSTERS	0x0002	// blocks monsters only
#define ML_TWOSIDED			0x0004	// backside will not be present at all if not two sided

#define ML_DONTPEGTOP		0x0008	// upper texture unpegged
#define ML_DONTPEGBOTTOM	0x0010	// lower texture unpegged
#define ML_SECRET			0x0020	// don't map as two sided: IT'S A SECRET!
#define ML_SOUNDBLOCK		0x0040	// don't let sound cross two of these
#define ML_DONTDRAW 		0x0080	// don't draw on the automap
#define ML_MAPPED			0x0100	// set if already drawn in automap
#define ML_REPEAT_SPECIAL	0x0200	// special is repeatable
#define ML_SPAC_SHIFT		10
#define ML_SPAC_MASK		0x1c00
#define GET_SPAC(flags)		((flags&ML_SPAC_MASK)>>ML_SPAC_SHIFT)

// Special activation types
#define SPAC_CROSS		0	// when player crosses line
#define SPAC_USE		1	// when player uses line
#define SPAC_MCROSS		2	// when monster crosses line
#define SPAC_IMPACT		3	// when projectile hits line
#define SPAC_PUSH		4	// when player/monster pushes line
#define SPAC_PCROSS		5	// when projectile crosses line
#define SPAC_USETHROUGH	6	// SPAC_USE, but passes it through

// [RH] Monsters (as well as players) can active the line
#define ML_MONSTERSCANACTIVATE		0x2000

// [RH] BOOM's ML_PASSUSE flag (conflicts with ML_REPEATSPECIAL)
#define ML_PASSUSE_BOOM				0x0200

// [RH] Line blocks everything
#define ML_BLOCKEVERYTHING			0x8000


/*
   this data structure contains the information about the SIDEDEFS
   */
struct SideDef
{
    short xoff;      /* X offset for texture */
    short yoff;      /* Y offset for texture */
    char tex1[8];  /* texture name for the part above */
    char tex2[8];  /* texture name for the part below */
    char tex3[8];  /* texture name for the regular part */
    short sector;    /* adjacent sector  */
};
typedef struct SideDef *SDPtr;



/*
   this data structure contains the information about the VERTEXES
   */
struct Vertex
{
    short x;         /* X coordinate */
    short y;         /* Y coordinate */
};
typedef struct Vertex *VPtr;



/*
   this data structure contains the information about the SEGS
   */
typedef struct Seg *SEPtr;
struct Seg
{
    SEPtr next;      /* next Seg in list */
    short start;     /* from this vertex ... */
    short end;       /* ... to this vertex */
    short angle;    /* angle (0 = east, 16384 = north, ...) */
    short linedef;   /* linedef that this seg goes along*/
    short flip;      /* true if not the same direction as linedef */
    unsigned short dist;     /* distance from starting point */
};



/*
   this data structure contains the information about the SSECTORS
   */
typedef struct SSector *SSPtr;
struct SSector
{
    SSPtr next;	  /* next Sub-Sector in list */
    short num;       /* number of Segs in this Sub-Sector */
    unsigned short first;     /* first Seg */
};



/*
   this data structure contains the information about the NODES
   */
typedef struct Node *NPtr;
struct Node
{
    short x, y;                         /* starting poshort */
    short dx, dy;                       /* offset to ending point */
    short miny1, maxy1, minx1, maxx1;   /* bounding rectangle 1 */
    short miny2, maxy2, minx2, maxx2;   /* bounding rectangle 2 */
    short child1, child2;               /* Node or SSector (if high bit is set) */
    NPtr node1, node2;                /* pointer if the child is a Node */
    short num;                          /* number given to this Node */
};



/*
   this data structure contains the information about the SECTORS
   */
struct Sector
{
    short floorh;    /* floor height */
    short ceilh;     /* ceiling height */
    char floort[8];/* floor texture */
    char ceilt[8]; /* ceiling texture */
    short light;     /* light level (0-255) */
    short special;   /* special behaviour (0 = normal, 9 = secret, ...) */
    short tag;       /* sector activated by a linedef with the same tag */
};
typedef struct Sector *SPtr;


struct Behavior
{
	char magic[4];	/* 'ACS\0' */
	int tableOffs;
	int data[1];
};
typedef struct Behavior *BPtr;



/* SO: added these 2 25/02/95 to speed up displaying wall textures */

typedef struct {
    char name[8]; /* name of this texture */
	unsigned short flags;	/* [CR] texture flags */
    BYTE scalex;  /* [RH] texture scalars */
    BYTE scaley;
    unsigned short width;  /* dimensions of this texture */
    unsigned short height;
    short zero3;
    short zero4;
    short nPatches; /* number of patch descriptors following */
} Texture;
#define TF_SCALE_OFFSET	0x8000

typedef struct {
    short xOffset,
	yOffset,
	PatchNumber,
	One,
	Zero;
} PatchDesc;

/* these are to make drawing pictures easier */
typedef struct {
	short width, height, xofs, yofs;
} PHeader;

typedef struct {
	BYTE yofs, npixels;
} PColumn;

#endif
/* _WSTRUCTS_H */

/* end of file */
