/*
   DETH - Doom Editor for Total Headcases, by Simon Oke and Antony Burden.
   HETH - Hexen Editor for Total Headcases, by Antony Burden.
   
   You are allowed to use any parts of this code in another program, as
   long as you give credits to the authors in the documentation and in
   the program itself.  Read the file README.1ST for more information.
   
   This program comes with absolutely no warranty.

   THINGS.C - Thing name and type routines.
   */

/* the includes */
#include <assert.h>
#include "deu.h"
#include "levels.h"

#define _WHITE	15

thing_class *Thing_classes;
spawn_class *Spawn_classes;
scheme_class *Scheme_classes;

BCINT TIDList[65536];

typedef struct {
	BCINT type;
	thing_type *thing;
} thing_arr;

typedef struct {
	BCINT type;
	scheme_type *scheme;
} scheme_arr;

typedef struct {
	BCINT type;
	spawn_type *spawn;
} spawn_arr;

thing_arr *Thing_type_array;
int maxThing;

scheme_arr *Scheme_type_array;
int maxScheme;

spawn_arr *Spawn_type_array;
int maxSpawn;

int BCINT_lessp(const void *x, const void *y)
{
	return (int)(*((BCINT *)x) - *((BCINT *)y));
}

void index_things()
{
	thing_class *c;
	thing_type *t;
	thing_arr *p;
	int nThingTypes = 0;
	
	for(c = Thing_classes; c; c = c->next)
		for(t = c->types; t; t = t->next)
			nThingTypes++;
	
	p = Thing_type_array = (thing_arr *)myalloc(nThingTypes * sizeof(thing_arr));
//	assert(Thing_type_array);
	
	for(c = Thing_classes; c; c = c->next)
	{
		for(t = c->types; t; t = t->next)
		{
			p->type = t->type;
			p->thing = t;
			p++;
		}
	}
	
	qsort(Thing_type_array, nThingTypes, sizeof(thing_arr), BCINT_lessp);
	maxThing = nThingTypes - 1;
}

void index_schemes()
{
	scheme_class *c;
	scheme_type *t;
	scheme_arr *p;
	int nSchemeTypes = 0;
	
	for(c = Scheme_classes; c; c = c->next)
		for(t = c->types; t; t = t->next)
			nSchemeTypes++;
	
	p = Scheme_type_array = (scheme_arr *)myalloc(nSchemeTypes * sizeof(scheme_arr));
//	assert(Scheme_type_array);
	
	for(c = Scheme_classes; c; c = c->next)
		for(t = c->types; t; t = t->next) {
			p->type = t->type;
			p->scheme = t;
			p++;
		}
	
	qsort(Scheme_type_array, nSchemeTypes, sizeof(scheme_arr),
		  BCINT_lessp);
	maxScheme = nSchemeTypes - 1;
}


void index_spawns(void)
{
	spawn_class *c;
	spawn_type *t;
	spawn_arr *p;
	int nSpawnTypes = 0;
	
	for(c = Spawn_classes; c; c = c->next)
		for(t = c->types; t; t = t->next)
			nSpawnTypes++;
	
	p = Spawn_type_array = (spawn_arr *)myalloc(nSpawnTypes * sizeof(spawn_arr));
//	assert(Spawn_type_array);
	
	for(c = Spawn_classes; c; c = c->next)
		for(t = c->types; t; t = t->next) {
			p->type = t->type;
			p->spawn = t;
			p++;
		}
	
	qsort(Spawn_type_array, nSpawnTypes, sizeof(spawn_arr),
		  BCINT_lessp);
	maxSpawn = nSpawnTypes - 1;
}


/*
   get the colour of a thing
   */

int FindThingType (int type)
{
	static int last = -1, lasti = -1;
	int low = 0, high = maxThing;

	if (type == last)
	{
		return lasti;
	}

	do
	{
		int mid = (low + high) / 2;
		int t = Thing_type_array[mid].type;
		if (t == type)
		{
			last = type;
			lasti = mid;
			return mid;
		}
		else if (t > type)
		{
			high = mid - 1;
		}
		else
		{
			low = mid + 1;
		}
	} while (low <= high);
	return -1;
}

BCINT GetThingExtra (BCINT type)
{
	type = FindThingType (type);
	return type >= 0 ? Thing_type_array[type].thing->extra : 0;
}

BCINT GetThingColour( BCINT type)
{
	int thing = FindThingType (type);
	int col = (thing >= 0 ? Thing_type_array[thing].thing->colour : _WHITE);
	return col == _WHITE ? WHITE : col;
}

BCINT GetThingPathType (BCINT type)
{
	type = FindThingType (type);
	return type >= 0 ? Thing_type_array[type].thing->pathtype : 0;
}

/*
   get the name of a thing
   */

char *GetThingName( BCINT type)
{
	int thing = FindThingType (type);
	return thing >= 0 ? Thing_type_array[thing].thing->name : "Unknown thing type";
}

/*
   get the name of a spawnable thing
   */

char *GetSpawnName( BCINT arg2)
{
   	int lowbound = 0, highbound = maxSpawn, middle;
	BYTE t;
	
	do {
		middle = (lowbound + highbound) / 2;
		t = Spawn_type_array[middle].type;
		
		if(t == arg2)
			return Spawn_type_array[middle].spawn->name;
		
		if(t > arg2)
			highbound = middle - 1;
		else
			lowbound = middle + 1;
	} while(lowbound <= highbound);
    
    if (arg2 <= 0) 
		return("None");
    else
		return("ERROR!");
}

/*
   get the name of the sprite to show
   */
char *GetThingSprite( BCINT type)
{
	int thing = FindThingType (type);
	return thing >= 0 ? Thing_type_array[thing].thing->sprite : "";
}

/*
   get the name of the spawn sprite to show
   */


char *GetSpawnSprite( BCINT type)
{
   	int lowbound = 0, highbound = maxSpawn, middle;
	BCINT t;
	
	do {
		middle = (lowbound + highbound) / 2;
		t = Spawn_type_array[middle].type;
		
		if(t == type)
			return Spawn_type_array[middle].spawn->sprite;
		
		if(t > type)
			highbound = middle - 1;
		else
			lowbound = middle + 1;
	} while(lowbound <= highbound);
    
    /* fall through */
    return("");
}

/*
   get the size of a thing
   */
int GetThingRadius( int type, int arg0)
{
	int thing = FindThingType (type);
	if (thing < 0)
		return 16;

	if ((Thing_type_array[thing].thing->extra & TX_RESIZER) && arg0)
		return arg0;

	return Thing_type_array[thing].thing->radius;
}

/* [RH] Get thing argument descriptions */

char *GetThingArgs( BCINT type)
{
	int thing = FindThingType (type);
	if (thing < 0 || Thing_type_array[thing].thing->args[0] == ' ')
		return NULL;
	return Thing_type_array[thing].thing->args;
}

char **GetThingArgText( BCINT type)
{
	static char *error[5] = { "ERROR", "ERROR", "ERROR", "ERROR", "ERROR" };
	int thing = FindThingType (type);
	return thing >= 0 ? Thing_type_array[thing].thing->argtext : error;
}

char **GetThingEnums (BCINT type)
{
	static char *error[5] = { 0 };
	int thing = FindThingType (type);
	return thing >= 0 ? Thing_type_array[thing].thing->enums : error;
}

/*
   get the name of the angle
   */

char *GetAngleName( BCINT angle)
{
    switch (angle) {
    case 0:
		return "(East)";
    case 45:
		return "(NorthEast)";
    case 90:
		return "(North)";
    case 135:
		return "(NorthWest)";
    case 180:
		return "(West)";
    case 225:
		return "(SouthWest)";
    case 270:
		return "(South)";
    case 315:
		return "(SouthEast)";
    }
    return "Other";
}


int SetThingFlagFilter(int old)
{
	int new = old;
	/* war == peace */
	
	int selection;

	MenuDefault = 0;
	do {
		selection = DisplayMenu(-1, -1, "Set Filter for Things",
								(new & TF_EASY) ? "* Easy" : "  Easy",
								(new & TF_MEDIUM) ? "* Medium" : "  Medium",
								(new & TF_HARD) ? "* Hard" : "  Hard",
								(new & TF_FIGHTER) ? "* Fighter" : "  Fighter",
								(new & TF_CLERIC) ? "* Cleric" : "  Cleric",
								(new & TF_MAGE) ? "* Mage" : "  Mage",
								(new & TF_SINGLE) ? "* Single" : "  Single",
								(new & TF_COOP) ? "* Cooperative" : "  Cooperative",
								(new & TF_DEATHM) ? "* Death Match" : "  Death Match",
								"  Cancel", NULL);

		switch(selection) {
		case 0: return new;
		case 1:	new ^= TF_EASY; break;
		case 2:	new ^= TF_MEDIUM; break;
		case 3:	new ^= TF_HARD; break;
		case 4:	new ^= TF_FIGHTER; break;
		case 5:	new ^= TF_CLERIC; break;
		case 6:	new ^= TF_MAGE; break;
		case 7:	new ^= TF_SINGLE; break;
		case 8:	new ^= TF_COOP; break;
		case 9:	new ^= TF_DEATHM; break;
		case 10:	return old;
		}
		MenuDefault = selection - 1;
	} while(1);
}

void InitTIDList (void)
{
	int tid;

	for (tid = 0; tid < 65536; ++tid)
	{
		TIDList[tid] = -1;
	}
}

void UnlinkTID (int thing)
{
	int tid = Things[thing].tid;

	if (tid != 0)
	{
		if (Things[thing].prevByID >= 0)
		{
			Things[Things[thing].prevByID].nextByID = Things[thing].nextByID;
		}
		else
		{
			TIDList[tid] = Things[thing].nextByID;
		}
		if (Things[thing].nextByID >= 0)
		{
			Things[Things[thing].nextByID].prevByID = Things[thing].prevByID;
		}
	}
}

void LinkTID (int thing)
{
	int tid = Things[thing].tid;
	int i, j;

	if (tid != 0)
	{
		// Insert the thing into the list in order
		for (j = -1, i = TIDList[tid]; i >= 0 && i < thing; j = i, i = Things[i].nextByID)
			;
		Things[thing].prevByID = j;
		Things[thing].nextByID = i;
		if (j >= 0)
		{
			Things[j].nextByID = thing;
		}
		else
		{
			TIDList[tid] = thing;
		}
		if (i >= 0)
		{
			Things[i].prevByID = thing;
		}
	}
	else
	{
		Things[thing].nextByID = Things[thing].prevByID = -1;
	}
}

void FixTIDList (void)
{
	int i;

	InitTIDList ();
	for (i = 0; i < NumThings; ++i)
	{
		LinkTID (i);
	}
}

void SetThingTID (int thing, int tid)
{
	UnlinkTID (thing);
	Things[thing].tid = tid;
	LinkTID (thing);
}

int FindThingByTID (int type, int tid)
{
	int i;

	if (type == 0)
	{
		return TIDList[tid];
	}

	for (i = TIDList[tid]; i >= 0; i = Things[i].nextByID)
	{
		if (Things[i].type == type)
		{
			return i;
		}
	}
	return -1;
}

/*
   find a free tid number
   */

BCINT FindFreeTid() /* SWAP! */
{
	int tid;

	for (tid = 1; tid < 65536; ++tid)
	{
		if (TIDList[tid] < 0)
			return tid;
	}
	return 0;
}



/* end of file */
